/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.reporters.pdd;

import com.ymock.util.Logger;

// URL management
import java.net.URL;

// collection management
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

// JAXB
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.XmlValue;

// for string manipulations
import org.apache.commons.lang.StringUtils;

/**
 * One work package in the WBS.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: WorkPackage.java 1694 2011-07-21 19:26:08Z guard $
 * @see Wbs
 */
@XmlType(propOrder = { "issue", "subject", "body", "dependants" })
public final class WorkPackage {

    /**
     * Associative array of children.
     */
    private final Map<String, WorkPackage> dependants =
        new HashMap<String, WorkPackage>();

    /**
     * Holder of this WP, the parent WP. Actually this link is
     * used only for {@link getMarker()} calculation and is necessary
     * only during XML marshaling in {@link toXml()}.
     * @see #WorkPackage()
     */
    private WorkPackage parent;

    /**
     * Estimate in minutes.
     * @see #setEstimate(Integer)
     * @see #getEstimate()
     */
    private Integer estimate;

    /**
     * Issue information in XML.
     * @see #setIssue(String, URL, String)
     * @see #getIssue()
     */
    private XmlIssue issue;

    /**
     * Subject of the WP.
     * @see #setSubject(String)
     * @see #getSubject()
     */
    private String subject;

    /**
     * Body of the WP.
     * @see #appendBody(String)
     * @see #getBody()
     */
    private String body = "";

    /**
     * Public ctor.
     * @see Wbs
     */
    public WorkPackage() {
        this.parent = null;
    }

    /**
     * Private ctor.
     * @param prnt Parent WP
     * @see #locate(String)
     */
    private WorkPackage(final WorkPackage prnt) {
        this.parent = prnt;
    }

    /**
     * Get list of dependants.
     * @return The dependants
     */
    @XmlElementWrapper(name = "dependants")
    @XmlElement(name = "wp")
    public Collection<WorkPackage> getDependants() {
        return this.dependants.values();
    }

    /**
     * Get name/marker of the package. Internal error (the exception)
     * should never happen here.
     * @return The name of it (unique marker)
     */
    @XmlAttribute(name = "marker")
    public String getMarker() {
        if (this.parent == null) {
            return "SuD";
        }
        for (Map.Entry<String, WorkPackage> entry
            : this.parent.dependants.entrySet()) {
            if (entry.getValue() == this) {
                return entry.getKey();
            }
        }
        throw new IllegalStateException("Internal error");
    }

    /**
     * Get estimate of the WP in minutes.
     * @return The estimate, in minutes
     */
    @XmlAttribute(name = "estimate")
    public Integer getEstimate() {
        return this.estimate;
    }

    /**
     * What this WP is about?
     * @return Short description of the WP
     */
    @XmlElement(name = "subject")
    public String getSubject() {
        return this.subject;
    }

    /**
     * What this WP is about, full description.
     * @return Full description of the WP
     */
    @XmlElement(name = "description")
    public String getBody() {
        return this.body;
    }

    /**
     * XML Issue.
     */
    @XmlType
    public static final class XmlIssue {
        /**
         * Name of the issue.
         */
        private String name;
        /**
         * URL of the issue.
         */
        private URL url;
        /**
         * Name of assignee.
         */
        private String assignee;
        /**
         * Public ctor, required by JAXB marshalling.
         */
        public XmlIssue() {
            // intentionally empty
        }
        /**
         * Public ctor.
         * @param pname The name of issue
         * @param purl The url of issue
         * @param passignee The name of assignee
         * @see WorkPackage#setIssue(WorkPackage.XmlIssue)
         */
        public XmlIssue(final String pname, final URL purl,
            final String passignee) {
            this.name = pname;
            this.url = purl;
            this.assignee = passignee;
        }
        /**
         * Get name of the issue.
         * @return The name
         */
        @XmlValue
        public String getName() {
            return this.name;
        }
        /**
         * Get URL of the issue.
         * @return The URL
         */
        @XmlAttribute(name = "href")
        public URL getUrl() {
            return this.url;
        }
        /**
         * Get assignee of the issue.
         * @return The assignee
         */
        @XmlAttribute(name = "assignee")
        public String getAssignee() {
            return this.assignee;
        }
    }

    /**
     * Get name of the issue, assigned to this WP.
     * @return The name
     */
    @XmlElement(name = "issue")
    public XmlIssue getIssue() {
        return this.issue;
    }

    /**
     * This WP is assigned (has issue)?
     * @return Is it?
     */
    public boolean isAssigned() {
        return this.issue != null;
    }

    /**
     * Get estimate of the WP in minutes.
     * @param mins The estimate
     * @return This
     */
    public WorkPackage setEstimate(final Integer mins) {
        this.estimate = mins;
        return this;
    }

    /**
     * What this WP is about?
     * @param subj The subject
     * @return This
     */
    public WorkPackage setSubject(final String subj) {
        this.subject = subj;
        return this;
    }

    /**
     * Append new text to the body.
     * @param text The text
     * @return This
     */
    public WorkPackage appendBody(final String text) {
        if (this.body.length() > 0) {
            this.body = this.body + "\n";
        }
        this.body = this.body + text;
        return this;
    }

    /**
     * Set the assigned issue.
     * @param pissue The issue
     * @return This
     */
    public WorkPackage setIssue(final XmlIssue pissue) {
        this.issue = pissue;
        return this;
    }

    /**
     * Optimize all WPs inside, and place them to the
     * right places. Every dependant in this work package
     * is going to be placed somewhere beneath, in one of
     * sub-packages, if possible. When such a replacement
     * takes place the original WP is destroyed, and all its
     * dependants are migrated to the new place found.
     */
    public void optimize() {
        // Associative array, where key is a name of the
        // issue associated and a value is the owner of this
        // association.
        final Map<String, WorkPackage> places =
            new HashMap<String, WorkPackage>();
        // Recursively discover all places possible, in this
        // object and all dependants down.
        this.discover(places);
        // Now let's try to move our dependants to the
        // sub-deps.
        final Set<String> migrated = new HashSet<String>();
        for (String dep : this.dependants.keySet()) {
            if (places.containsKey(dep)) {
                places.get(dep).migrate(this.dependants.get(dep));
                migrated.add(dep);
            }
        }
        for (String toRemove : migrated) {
            this.dependants.remove(toRemove);
        }
        Logger.debug(
            this,
            "#optimize(): migrated [%s]",
            StringUtils.join(migrated, ", ")
        );
    }

    /**
     * Migrate all dependants from the provided WP into this
     * work package.
     * @param dep The dependant
     * @see #optimize()
     */
    public void migrate(final WorkPackage dep) {
        for (WorkPackage toMove : dep.dependants.values()) {
            this.dependants.put(toMove.getMarker(), toMove);
        }
    }

    /**
     * Find sub-package by its name/marker. If the work package
     * is not found we should create it.
     * @param marker The marker to look for, recursively
     * @return Found work package, or created one
     */
    public WorkPackage locate(final String marker) {
        WorkPackage found = this.find(marker);
        if (found == null) {
            found = new WorkPackage(this);
            this.dependants.put(marker, found);
        }
        return found;
    }

    /**
     * Find sub-package by its name/marker. If the work package
     * is not found we return NULL. This method recursively calls
     * itself.
     * @param marker The marker to look for, recursively
     * @return Found work package or NULL
     */
    private WorkPackage find(final String marker) {
        WorkPackage found = null;
        if (this.dependants.containsKey(marker)) {
            found = this.dependants.get(marker);
        } else {
            for (Map.Entry<String, WorkPackage> pair
                : this.dependants.entrySet()) {
                found = pair.getValue().find(marker);
                if (found != null) {
                    break;
                }
            }
        }
        return found;
    }

    /**
     * Discover all places in this WP and all its
     * dependants (recursively).
     * @param places Associative array where key is the name of issue
     * @see #optimize()
     */
    private void discover(final Map<String, WorkPackage> places) {
        for (WorkPackage wp : this.dependants.values()) {
            if (wp.isAssigned()) {
                places.put(
                    wp.getIssue().getName(),
                    wp
                );
            }
            wp.discover(places);
        }
    }

}
